import java.util.*;
import javax.vecmath.*;
import java.lang.Math;


/*
 * Attributes needed :
 * Vertex :
 * - Its new value
 * - If its odd or even
 * Face :
 * - odd vertex
 * - If its odd or even
 * - If its boundary or not
 * - has it been edge flipped or not
 * Edge :
 * - 2 vertices (if its a boundary edge)
 */
public class Root3Subdivision implements SubdivisionAlgorithm
{
    int oddVertexCounter = 0;
    public class VertexAttributes
    {
        Point3d newValue = null;
        //isOdd is used to determine the edge that connects two even vertices
        //This is used mainly for edge flipping.  Should be set during 1-3
        //split
        boolean isOdd = false;
        double hashCode;
    }

    //Note : no FaceAttributes on a face implies that it is even
    public class FaceAttributes
    {
        Point3d faceVertex = null;
        //A face is considered even if it was a result of an edge flip
        //it is considered odd if it was a result of a 1-3 split
        //If a face is odd it can NOT undergo another 1-3 split.
        boolean isOdd;
        //Used to determine if the face has gone under edge flipping or not
        boolean hasFlipped = false;

        int hashCode = 0;
    }
    public class EdgeAttributes
    {
        Point3d v1 = null;
        Point3d v2 = null;

        int v1HashCode = 0;
        int v2HashCode = 0;
    }

    private static final double FOUR = 4.0;
    private static final double TEN = 10.0;
    private static final double SIXTEEN = 16.0;
    private static final double NINETEEN = 19.0;
    private static final double ONE_TWENTY_SEVENTH = (1.0/27.0);
    private static final double ONE_THIRD = (1.0/3.0);
    /*
    private static final double ONE_EIGHTH = 0.125;
    private static final double THREE_EIGHTH = 0.375;
    private static final double SIX_EIGHTH = 0.75;
    private static final double ONE_FOURTH = 0.25;
    private static final double ONE_HALF = 0.5;
    */

    private  Map alphaNCache = new HashMap();


    private boolean hasCalculated(VertexHandle vh)
    {
        VertexAttributes va =
            (VertexAttributes)vh.getAttribute();
        if (va == null || va.newValue == null) {
            return false;
        }
        return true;
    }

    private  double getAlphaN(int valence) 
    {
        Integer key = new Integer(valence);
        Double ret = (Double)alphaNCache.get(key);
        if (ret == null) {
            ret = 
                new Double((4.0 - 
                        Math.cos((2.0 * Math.PI)/(double)valence) * 2.0) /
                        9.0);
            alphaNCache.put(key, ret);
        }
        return ret.doubleValue();
    }

    private  Point3d calc3iMinus1(Point3d PiMinus1, Point3d Pi, 
            Point3d PiPlus1)
    {
        Point3d ret = new Point3d(PiMinus1);
        Point3d tempPoint;

        ret.scale(TEN);
        tempPoint = new Point3d(Pi);
        tempPoint.scale(SIXTEEN);
        ret.add(tempPoint);
        tempPoint = new Point3d(PiPlus1);
        ret.add(tempPoint);
        
        ret.scale(ONE_TWENTY_SEVENTH);
        return ret;

    }

    private  Point3d calc3i(Point3d PiMinus1, Point3d Pi, 
            Point3d PiPlus1)
    {
        Point3d ret = new Point3d(PiMinus1);
        Point3d tempPoint;

        ret.scale(FOUR);
        tempPoint = new Point3d(Pi);
        tempPoint.scale(NINETEEN);
        ret.add(tempPoint);
        tempPoint = new Point3d(PiPlus1);
        tempPoint.scale(FOUR);
        ret.add(tempPoint);
        
        ret.scale(ONE_TWENTY_SEVENTH);
        return ret;

    }

    private  Point3d calc3iPlus1(Point3d PiMinus1, Point3d Pi, 
            Point3d PiPlus1)
    {
        Point3d ret = new Point3d(PiMinus1);
        Point3d tempPoint;

        tempPoint = new Point3d(Pi);
        tempPoint.scale(SIXTEEN);
        ret.add(tempPoint);
        tempPoint = new Point3d(PiPlus1);
        tempPoint.scale(TEN);
        ret.add(tempPoint);
        
        ret.scale(ONE_TWENTY_SEVENTH);
        return ret;
    }

    /**
     * Swap the edges of the two faces bordering the evenEdge
     * If evenEdge doesn't have a sym then the original face is added
     * into the newMesh
     */
    private void swapEdge(Mesh newMesh, Mesh oldMesh, HalfEdge evenEdge)
    {
        //we can only do an edge flip if 
        //we HAVE a neighbour to edge flip with
        VertexHandle []newVertexHandle = new VertexHandle[4];
        VertexHandle []newFace = new VertexHandle[3];
        if (evenEdge.sym != null) {
            //First we want to mark the face we're swapping edges with
            //that we did the swap already so we wouldn't add faces 
            //twice
            FaceAttributes fa = 
                (FaceAttributes)evenEdge.sym.getFace().getAttribute();
            fa.hasFlipped = true;
            evenEdge.sym.getFace().setAttribute(fa);

            //Now we move onto actually doing the edge flip
            newVertexHandle[0] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.getVertexHashCode()).getVertex(),
                        evenEdge.getVertexHashCode());
            newVertexHandle[1] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.next.getVertexHashCode()).getVertex(),
                        evenEdge.next.getVertexHashCode());
            newVertexHandle[2] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.next.next.getVertexHashCode()).getVertex(),
                        evenEdge.next.next.getVertexHashCode());
            newVertexHandle[3] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.sym.next.getVertexHashCode()).getVertex(),
                        evenEdge.sym.next.getVertexHashCode());

            newFace[0] = newVertexHandle[2];
            newFace[1] = newVertexHandle[3];
            newFace[2] = newVertexHandle[1];
            //No face attributes needed since no face attributes
            //implies an even face
            newMesh.addFace(newFace);

            newFace[0] = newVertexHandle[3];
            newFace[1] = newVertexHandle[0];
            newFace[2] = newVertexHandle[1];
            //No face attributes needed since no face attributes
            //implies an even face
            newMesh.addFace(newFace);
        }//if the face was NOT a boundary face
        else {
            //This face doesn't have a neighbour to swap edges with
            //so just add it as is
            newVertexHandle[0] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.getVertexHashCode()).getVertex(),
                        evenEdge.getVertexHashCode());
            newVertexHandle[1] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.next.getVertexHashCode()).getVertex(),
                        evenEdge.next.getVertexHashCode());
            newVertexHandle[2] = 
                newMesh.addVertex(
                        oldMesh.getVertexHandle(
                            evenEdge.next.next.getVertexHashCode()).getVertex(),
                        evenEdge.next.next.getVertexHashCode());

            newFace[0] = newVertexHandle[0];
            newFace[1] = newVertexHandle[1];
            newFace[2] = newVertexHandle[2];
            FaceAttributes fa = new FaceAttributes();
            //There is ONE more special case.  If we're passed a face that HAS
            //no even edges, then we add the same face BUT treat it as an
            //even face so that it will get a 1-3 split in the next step
            VertexAttributes va =
                (VertexAttributes) 
                (oldMesh.getVertexHandle(
                                evenEdge.getVertexHashCode()).getAttribute());
            fa.isOdd = !va.isOdd;
            newMesh.addFace(newFace, fa);
        }//if the face was a boundary face
    }

    //Given a face f in oldMesh that has been prepared by 
    //prepareThreeToOneSplit, actually DO the 3-1 split and insert the
    //result in newMesh
    private void threeToOneSplit(Mesh oldMesh, Mesh newMesh,
            Face oldFace)
    {
        Iterator edgeIt = oldFace.halfEdgeIterator();
        VertexHandle[] newEvenHandles = new VertexHandle[3];
        VertexHandle[] newOddHandles = new VertexHandle[2];
        VertexHandle[] newFace = new VertexHandle[3];
        int index = 0;
        int boundaryIndex = 0;
        boolean isBoundary = false;
        FaceAttributes fa = 
            (FaceAttributes)oldFace.getAttribute();

        while (edgeIt.hasNext()) {
            HalfEdge he = (HalfEdge)edgeIt.next();
            //First see if we need to do the special boundary split
            if ((he.sym == null) && (fa.isOdd)) {
                //System.out.println("boundary rule applied");
                //boundary edge found
                isBoundary = true;
                EdgeAttributes ea = 
                    (EdgeAttributes)he.getAttribute();

                //Set the odd vertex handles to be the boundary
                //vertices
                VertexAttributes va =
                    new VertexAttributes();
                va.isOdd = true;
                newOddHandles[0] = newMesh.addVertex(ea.v1, ea.v1HashCode);
                newOddHandles[0].setAttribute(va);
                va =
                    new VertexAttributes();
                va.isOdd = true;
                newOddHandles[1] = newMesh.addVertex(ea.v2, ea.v2HashCode);
                newOddHandles[1].setAttribute(va);
                boundaryIndex = index;
            }//if this is a boundary edge

            //Take care of the even vertices
            VertexAttributes oldVa = 
                (VertexAttributes)oldMesh.getVertexHandle(
                                                          he.getVertexHashCode()).getAttribute();

            VertexAttributes va = 
                new VertexAttributes();
            va.isOdd = false;
            newEvenHandles[index] =
                newMesh.addVertex(oldVa.newValue, he.getVertexHashCode());
            newEvenHandles[index].setAttribute(va);
            index++;
        }//for all edges in this old face

        //Now insert the 1-3 split triangles into the new mesh
        //If this is an odd face AND a boundary face, we do the special split
        if (isBoundary) {
            //Insert the 3 new triangles into the new mesh
            int prevBoundaryIndex = 
                (boundaryIndex - 1 < 0 ? 2 : boundaryIndex - 1);
            int nonBoundaryIndex = 
                ((boundaryIndex + 1) % 3);
            fa = null;

            newFace[0] = newEvenHandles[prevBoundaryIndex];
            newFace[1] = newOddHandles[0];
            newFace[2] = newEvenHandles[nonBoundaryIndex];
            fa = new FaceAttributes();
            fa.isOdd = true;
            newMesh.addFace(newFace, fa);

            newFace[0] = newOddHandles[0];
            newFace[1] = newOddHandles[1];
            newFace[2] = newEvenHandles[nonBoundaryIndex];
            fa = new FaceAttributes();
            fa.isOdd = true;
            newMesh.addFace(newFace, fa);

            newFace[0] = newOddHandles[1];
            newFace[1] = newEvenHandles[boundaryIndex];
            newFace[2] = newEvenHandles[nonBoundaryIndex];
            fa = new FaceAttributes();
            fa.isOdd = true;
            newMesh.addFace(newFace, fa);
        }
        else {
            //The face is an even face
            //Extract the odd vertex from the face attribute
            fa = (FaceAttributes)oldFace.getAttribute(); 
            newOddHandles[0] = 
                newMesh.addVertex(fa.faceVertex, fa.hashCode);
            VertexAttributes va = 
                new VertexAttributes();
            va.isOdd = true;
            newOddHandles[0].setAttribute(va);

            newFace[0] = newEvenHandles[0];
            newFace[1] = newEvenHandles[1];
            newFace[2] = newOddHandles[0];
            fa = new FaceAttributes();
            fa.isOdd = true;
            newMesh.addFace(newFace, fa);

            newFace[0] = newEvenHandles[1];
            newFace[1] = newEvenHandles[2];
            newFace[2] = newOddHandles[0];
            fa = new FaceAttributes();
            fa.isOdd = true;
            newMesh.addFace(newFace, fa);

            newFace[0] = newEvenHandles[2];
            newFace[1] = newEvenHandles[0];
            newFace[2] = newOddHandles[0];
            fa = new FaceAttributes();
            fa.isOdd = true;
            newMesh.addFace(newFace, fa);
        }
    }

    /**
     * If f is not a boundary polygon or if its even, then a 1-3 split
     * will be done on it, otherwise the even vertices are relaxed, and the
     * boundary edge will have two vertices added to it
     * Note that the actual 3-1 split is not performed, but is "prepared"
     */
    private void prepareThreeToOneSplit(Mesh oldMesh, Face f)
    {
        FaceAttributes faceAttributes = (FaceAttributes)f.getAttribute();
        if (f.isBoundary() && faceAttributes != null && 
                faceAttributes.isOdd) {
            //System.out.println("applying the special boundary mask");
            //apply the normal 3-1 split IF this face is even, but if its
            //odd, we apply the boundary mask and split the boundary edge 
            //into 3
            Iterator edgeIt = f.halfEdgeIterator();
            HalfEdge boundaryEdge = null;
            HalfEdge prevBoundaryEdge = null;
            while (edgeIt.hasNext()) {
                boundaryEdge = (HalfEdge)edgeIt.next();
                if (boundaryEdge.sym == null) {
                    break;
                }
            }
            //Sorta hackish, but yeah we can take shortcuts if we assume
            //this is a triangle mesh (which we can)
            prevBoundaryEdge = boundaryEdge.next.next;

            //We need a total of FOUR points
            //prevPoint is the point that points to the boundaryEdge
            //nextPoint is the point that the boundaryEdge points to
            //prevPrevPoint is the point that points to a boundaryEdge that
            //points to prevPoint
            //nextNextPoint is the point that is pointed to by the 
            //boundaryEdge comming out of nextPoint
            VertexHandle prevVertexHandle =
                oldMesh.getVertexHandle(
                        prevBoundaryEdge.getVertexHashCode());
            VertexHandle nextVertexHandle =
                oldMesh.getVertexHandle(
                        boundaryEdge.getVertexHashCode());
            Point3d prevPoint = new Point3d(prevVertexHandle.getVertex());
            Point3d nextPoint = new Point3d(nextVertexHandle.getVertex());
            Point3d prevPrevPoint = null;
            Point3d nextNextPoint = null;

            Iterator heIt = prevVertexHandle.oHalfEdgeIterator();
            //Set prevPrev
            while (heIt.hasNext()) {
                HalfEdge oHE = (HalfEdge)heIt.next();
                if (oHE.next.next.sym == null) {
                    prevPrevPoint = 
                        new Point3d(oldMesh.getVertexHandle(
                                    oHE.next.vertexHashCode).getVertex());
                    break;
                }
            }
            heIt = nextVertexHandle.oHalfEdgeIterator();
            //Set nextNext point
            while (heIt.hasNext()) {
                HalfEdge oHE = (HalfEdge)heIt.next();
                if (oHE.sym == null) {
                    nextNextPoint = 
                        new Point3d(oldMesh.getVertexHandle(
                                    oHE.vertexHashCode).getVertex());
                    break;
                }
            }

            /*
            System.out.println("prevPoint - " + prevPoint);
            System.out.println("nextPoint - " + nextPoint);
            System.out.println("prevPrevPoint - " + prevPrevPoint);
            System.out.println("nextNextPoint - " + nextNextPoint);
            */

            //Now that all the prep work is done, we actually START
            //applying the mask :(
            //NOTE : we check the first and second P'3i to see if they've 
            //already been calculated
            VertexAttributes tempVertexAttribute = null;
            Point3d tempPoint;
            //First P'3i
            VertexAttributes va = 
                new VertexAttributes();
            va.newValue = calc3i(prevPrevPoint, prevPoint, nextPoint);
            va.isOdd = false;
            prevVertexHandle.setAttribute(va);
            //Second P'3i
            va = 
                new VertexAttributes();
            va.newValue = calc3i(prevPoint, nextPoint, nextNextPoint);
            va.isOdd = false;
            nextVertexHandle.setAttribute(va);
            /*
            System.out.println("the new nextPoint is - " +
                    va.newValue);
                    */

            //Now do the two new edge points
            EdgeAttributes ea = 
                new EdgeAttributes();
            //P'3i+1
            ea.v1 = calc3iPlus1(prevPrevPoint, prevPoint, nextPoint);
            ea.v1HashCode = oddVertexCounter;
            oddVertexCounter++;
            //P'3i-1
            ea.v2 = calc3iMinus1(prevPoint, nextPoint, nextNextPoint);
            ea.v2HashCode = oddVertexCounter;
            oddVertexCounter++;
            boundaryEdge.setAttribute(ea);
            //FINALLY, we are done.. WHEW
        }
        else {
            //System.out.println("applying the normal mask");
            //apply the normal mask
            //FIXME : One thing I'm NOT sure is how to relax the boundary
            //vertices if the boundary face is an even face
            //For now I'm going to relax it like any other vertices
            //FIXME : do NOT relax a boundary vertex!!!!!
            //Another problem is that if a non-boundary face is asked to
            //relax a boundary vertex, what do we do?
            //The boundary face always have precedence
            Iterator vIt = f.vertexIterator();
            Point3d midPoint = new Point3d();
            while (vIt.hasNext()) {
                VertexHandle currentVertexHandle =
                    oldMesh.getVertexHandle(((Integer)vIt.next()).intValue());
                //Remember the value of currentVertexHandle to compute the
                //midpoint at the end
                midPoint.add(currentVertexHandle.getVertex());

                //If we've already relaxed this vertex then don't do it again
                //and just continue
                if (hasCalculated(currentVertexHandle)) {
                    //System.out.println("has calculated returned true");
                    continue;
                }

                Point3d newPoint = new Point3d();

                if (currentVertexHandle.isBoundary()) {
                    newPoint = new Point3d(currentVertexHandle.getVertex());
                }
                else {
                    Iterator oneRingIt = currentVertexHandle.vertexIterator();
                    int valence = 0;

                    while (oneRingIt.hasNext()) {
                        valence ++;
                        newPoint.add(
                                oldMesh.getVertexHandle(
                                    ((Integer)oneRingIt.next()).intValue()).
                                getVertex());
                    }//for all one-ring neighbours
                    //calculate the new even vertex
                    double alphaN =
                        getAlphaN(valence);
                    newPoint.scale((1.0/(double)valence) * (alphaN));
                    Point3d oldPoint = new Point3d(currentVertexHandle.getVertex());
                    oldPoint.scale(1.0 - alphaN);
                    newPoint.add(oldPoint);

                }

                //newPoint is the new value for the point at 
                //currentVertexHandle, so set it in the attributes
                VertexAttributes va =
                    new VertexAttributes();
                va.newValue = newPoint;
                va.isOdd = false;
                currentVertexHandle.setAttribute(va);
                
            }//for all vertices in the face
            //Lastly add the midpoint into this face
            FaceAttributes fa = new FaceAttributes();
            midPoint.scale(ONE_THIRD);
            fa.faceVertex = midPoint;
            fa.hashCode = oddVertexCounter;
            oddVertexCounter++;
            f.setAttribute(fa);
        }//if this polygon is not boundary OR even
    }

    /*
    private  Point3d applyBoundaryMask(Mesh oldMesh,
            VertexHandle currentVertexHandle)
    {
        Point3d newPoint =
            new Point3d(currentVertexHandle.getVertex());
        newPoint.scale(SIX_EIGHTH);
        Point3d tempPoint = null;

        Iterator heIt = currentVertexHandle.oHalfEdgeIterator();
        while (heIt.hasNext()) {
            HalfEdge oHE = (HalfEdge)heIt.next();
            if (oHE.sym == null) {
                tempPoint = 
                    new Point3d(oldMesh.getVertexHandle(
                                oHE.vertexHashCode).getVertex());
                tempPoint.scale(ONE_EIGHTH);
                newPoint.add(tempPoint);
            }

            if (oHE.next.next.sym == null) {
                tempPoint = 
                    new Point3d(oldMesh.getVertexHandle(
                                oHE.next.vertexHashCode).getVertex());
                tempPoint.scale(ONE_EIGHTH);
                newPoint.add(tempPoint);
            }
        }

        return newPoint;
    }

    private  Point3d applyBoundaryMaskWrong(Mesh oldMesh,
            VertexHandle currentVertexHandle)
    {
        VertexHandle tempVertexHandle;
        HalfEdge e;
        Point3d newPoint;
        Point3d tempPoint;

        //Use the boundary mask
        //(1/8)
        newPoint = new Point3d(currentVertexHandle.getVertex());
        newPoint.scale(ONE_EIGHTH);

        //e = currentHalfEdge.next;
        //We need to find which outgoing edge of this boundary vertex to
        //traverse.  Find one that will lead to the correct application 
        //of the boundary mask
        //Loop through all out-going edges, find one where next.sym
        //isn't null so we can apply the boundary mask
        Iterator it = currentVertexHandle.oHalfEdgeIterator();
        e = null;
        while (it.hasNext()) {
            e = (HalfEdge)it.next();
            if (e.next.sym != null) {
                break;
            }
            else {
                e = null;
            }
        }

        if (e == null) {
            //it is impossible to apply the boundary mask
            return newPoint;
        }

        //(3/8)
        tempVertexHandle = 
            oldMesh.getVertexHandle(e.getVertexHashCode());
        tempPoint = new Point3d(tempVertexHandle.getVertex());
        tempPoint.scale(THREE_EIGHTH);
        newPoint.add(tempPoint);

        //(3/8)
        e = e.next;
        tempVertexHandle = 
            oldMesh.getVertexHandle(e.getVertexHashCode());
        tempPoint = new Point3d(tempVertexHandle.getVertex());
        tempPoint.scale(THREE_EIGHTH);
        newPoint.add(tempPoint);

        //(1/8)
        tempPoint = 
            new Point3d(oldMesh.getVertexHandle(
                        e.sym.next.getVertexHashCode()).getVertex());
        tempPoint.scale(ONE_EIGHTH);
        newPoint.add(tempPoint);

        return newPoint;
    }
*/

    /*
    private  Point3d calcOddVertex(Mesh oldMesh,
            HalfEdge currentHalfEdge)
    {
        //First we have to find the 2 vertices the halfedge connects
        HalfEdge e = currentHalfEdge;
        while (e.next != currentHalfEdge) {
            e = e.next;
        }
        Point3d v1 = 
            new Point3d(oldMesh.getVertexHandle(
                        currentHalfEdge.getVertexHashCode()).getVertex());
        Point3d v2 = 
            new Point3d(oldMesh.getVertexHandle(
                        e.getVertexHashCode()).getVertex());

        //System.out.println("calcOddVertex() " + v1 + "-" + v2);

        Point3d newPoint;
        Point3d tempPoint;

        //If this edge is a boundary edge, then we apply the boundary
        //odd vertex rule
        if (currentHalfEdge.sym == null) {
            //(1/2)
            newPoint = new Point3d(v1);
            newPoint.scale(ONE_HALF);

            //(1/2)
            tempPoint = new Point3d(v2);
            tempPoint.scale(ONE_HALF);
            newPoint.add(tempPoint);

            //System.out.println("new odd vertex (at boundary) is " + newPoint);
            return newPoint;
        }

        //Not a boundary edge, continue normally
        //(3/8)
        newPoint = new Point3d(v1);
        //System.out.println("3/8 of " + newPoint);
        newPoint.scale(THREE_EIGHTH);
        //System.out.println("newPoint = " + newPoint);

        //(3/8)
        tempPoint = new Point3d(v2);
        //System.out.println("3/8 of " + tempPoint);
        tempPoint.scale(THREE_EIGHTH);
        newPoint.add(tempPoint);
        //System.out.println("newPoint = " + newPoint);

        //(1/8)
        tempPoint = 
            new Point3d(oldMesh.getVertexHandle(
                    currentHalfEdge.sym.next.getVertexHashCode()).getVertex());
        //System.out.println("1/8 of " + tempPoint);
        tempPoint.scale(ONE_EIGHTH);
        //System.out.println(" is " + tempPoint);
        newPoint.add(tempPoint);
        //System.out.println("newPoint = " + newPoint);

        //(1/8)
        tempPoint = 
            new Point3d(oldMesh.getVertexHandle(
                    currentHalfEdge.next.getVertexHashCode()).getVertex());
        //System.out.println("1/8 of " + tempPoint);
        tempPoint.scale(ONE_EIGHTH);
        //System.out.println(" is " + tempPoint);
        newPoint.add(tempPoint);
        //System.out.println("newPoint = " + newPoint);

        //System.out.println("new odd vertex is " + newPoint);
        return newPoint;
    }
*/

    /**
     * Calculates an even vertex 
     */
    /*
    private  Point3d calcEvenVertex(Mesh oldMesh,
            VertexHandle currentVertexHandle)
    {
        Point3d newPoint;
        
        //check to see if this is a boundary vertex
        if (currentVertexHandle.isBoundary()) {
            //System.out.println("boundary vertex detected");
            newPoint = applyBoundaryMask(oldMesh, currentVertexHandle);
        }
        else {
            //Use the normal mask
            //Get the "one-ring" of vertices
            Iterator oneRingIt = currentVertexHandle.vertexIterator();
            int valence = 0;
            newPoint = new Point3d();

            while (oneRingIt.hasNext()) {
                valence ++;
                newPoint.add(
                        oldMesh.getVertexHandle(oneRingIt.next()).getVertex());
            }//for all one-ring neighbours
            //calculate the new even vertex
            double alphaN =
                getAlphaN(valence);
            newPoint.scale((1.0/(double)valence) * (1.0 - alphaN));
            Point3d oldPoint = new Point3d(currentVertexHandle.getVertex());
            oldPoint.scale(alphaN);
            newPoint.add(oldPoint);
        }

        //System.out.println("new even vertex is " + newPoint);
        return newPoint;
    }
    */

    public Mesh subdivide(Mesh oldMesh)
        throws Exception
    {
        if (!oldMesh.isTriangular()) {
            throw new Exception("Root3Subdivision requires a triangular mesh");
        }

        oddVertexCounter = oldMesh.getLargestHashCode() + 1;
        Mesh newMesh = new Mesh();
        Iterator oldFaceIt = oldMesh.faceIterator();
        Face oldFace;
        Point3d[] oddVertex = new Point3d[3];
        Point3d[] evenVertex = new Point3d[3];
        HalfEdge currentHalfEdge;
        VertexHandle currentVertexHandle;
        int index;

        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            //******************************************************
            /*
            System.out.println("subdividing Face : ");
            Iterator vIt = oldFace.vertexIterator();
            int numVertex = 0;
            while (vIt.hasNext()) {
                VertexHandle vH = 
                    oldMesh.getVertexHandle(vIt.next());
                System.out.println("v - " + vH.getVertex());
                numVertex++;
            }
             */

            prepareThreeToOneSplit(oldMesh, oldFace);
            //TESTING____________________________________________
            //break;
        }//for all old faces

        //Now cycle through the old mesh again, "spawning" a new mesh
        oldFaceIt = oldMesh.faceIterator();
        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            threeToOneSplit(oldMesh, newMesh, oldFace);
        }//for all faces in the old mesh

        //Now cycle through the mesh, yet AGAIN, flipping the edges
        oldMesh = newMesh;
        newMesh = new Mesh();
        oldFaceIt = oldMesh.faceIterator();
        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            FaceAttributes fa = 
                (FaceAttributes)oldFace.getAttribute();
            //Only do edge flipping on those we havn't done it on
            if (!fa.hasFlipped) {
                //search for the edge that connects two even vertices
                Iterator heIt = oldFace.halfEdgeIterator();
                HalfEdge prevEdge = null;
                HalfEdge nextEdge = null;
                HalfEdge evenEdge = null;
                while (heIt.hasNext()) {
                    HalfEdge he = (HalfEdge)heIt.next();
                    VertexAttributes va =
                        (VertexAttributes)oldMesh.getVertexHandle(
                                he.getVertexHashCode()).getAttribute();
                    if (!va.isOdd) {
                        if (prevEdge == null) {
                            prevEdge = he;
                        }
                        else {
                            nextEdge = he;
                        }
                    }
                }//for all edges in this face
                if (prevEdge == null || nextEdge == null) {
                    //It is possible that this face has NO even edge
                    //in this case we do no edge swapping and just add it in
                    //System.out.println("NO even edge found");
                    if (prevEdge == null) {
                        evenEdge = nextEdge.next.next;
                    }
                    else {
                        evenEdge = prevEdge.next.next;
                    }
                }
                else {
                    if (prevEdge.next == nextEdge) {
                        evenEdge = nextEdge;
                    }
                    else {
                        evenEdge = prevEdge;
                    }
                }
                swapEdge(newMesh, oldMesh, evenEdge);
            }//flipped the edge of triangle oldFace ONLY if it wasn't already
        }//for all faces in the mesh after 1-3 split

        return newMesh;
    }
}
