import java.util.*;
import java.io.*;
import javax.vecmath.*;

public class Mesh
{
    //An index of all vertices in this map, index by their hashcode
    private Map vertexIndex;
    //An map of vertices to a list of all out-going edges
    private Map vertexToEdgeMap;
    //A list of all faces in the mesh
    private List faces;
    //An map of vertices to all their attributes
    private Map vertexToAttributeMap;
    private boolean isTriangular = true;
    private boolean triangulate;
    int largestHashCode = 0;

    public int getLargestHashCode()
    {
        return largestHashCode;
    }

    public Mesh()
    {
        this(false);
    }

    public Mesh(boolean triangulate)
    {
        vertexIndex = new HashMap();
        vertexToEdgeMap = new HashMap();
        faces = new LinkedList();
        vertexToAttributeMap = new HashMap();

        this.triangulate = triangulate;
    }

    public boolean isTriangular()
    {
        return isTriangular;
    }

    public int numFaces()
    {
        return faces.size();
    }

    public void toIV(File f)
        throws IOException
    {
        SortedMap vertexSet = new TreeMap();
        vertexSet.putAll(vertexIndex);
        //First write the header
        PrintWriter pw = new PrintWriter(new FileWriter(f));
        
        Map hashCodeToIndex = new HashMap();

        pw.println("#Inventor V2.0 ascii");
        pw.println("Material {");
        pw.println("    ambientColor  0.09 0.04 0.04");
        pw.println("    diffuseColor  0.9 0.4 0.4");
        pw.println("    specularColor  0.3 0.3 0.3");
        pw.println("    shininess 0.2");
        pw.println("    transparency   0");
        pw.println("}");

        pw.println("Separator {");
        //print all vertices
        pw.println("    Coordinate3 {");
        pw.println("            point   [");
        Iterator vertexIt = vertexSet.keySet().iterator();
        int index = 0;
        while(vertexIt.hasNext()) {
            VertexHandle vh = 
                getVertexHandle(((Integer)vertexIt.next()).intValue());
            hashCodeToIndex.put(new Integer(vh.getVertexHashCode()),
                    new Integer(index));
            index ++;
            Point3d v = vh.getVertex();
            pw.print("                " + v.x + " " + v.y + " " + v.z);
            if (vertexIt.hasNext()) {
                pw.println(",");
            }
            else {
                pw.println("");
            }
        }
        pw.println("            ]");
        pw.println("    }");
        
        //print all normals
        pw.println("    Normal {");
        pw.println("            vector   [");
        vertexIt = vertexSet.keySet().iterator();
        while(vertexIt.hasNext()) {
            VertexHandle vh = 
                getVertexHandle(((Integer)vertexIt.next()).intValue());
            Vector3d n = vh.calcNormal();
            pw.print("                " + n.x + " " + n.y + " " + n.z);
            if (vertexIt.hasNext()) {
                pw.println(",");
            }
            else {
                pw.println("");
            }
        }
        pw.println("            ]");
        pw.println("    }");
        
        //print all faces
        pw.println("    IndexedFaceSet {");
        pw.println("            materialIndex   -1");
        pw.println("            normalIndex     -1");
        pw.println("            textureCoordIndex       -1");
        pw.println("            coordIndex   [");
        Iterator fIt = faceIterator();
        while(fIt.hasNext()) {
            Iterator vIt = ((Face)fIt.next()).vertexIterator();
            pw.print("                                ");
            while (vIt.hasNext()) {
                int vindex =
                    ((Integer)hashCodeToIndex.get((Integer)vIt.next())).intValue();
                pw.print(vindex + ",");
            }//for all vertices in this face
            if (fIt.hasNext()) {
                pw.println("-1,");
            }
            else {
                pw.println("-1");
            }
        }//for all faces

        pw.println("            ]");
        pw.println("    }");

        pw.println("}");

        pw.close();
    }

    public void toS3D(File f)
        throws IOException
    {
        PrintWriter pw = new PrintWriter(new FileWriter(f));
        int numFaces = 0;
        
        Iterator fIt = faceIterator();
        while (fIt.hasNext()) {
            numFaces ++;
            List vertices = new LinkedList();
            List normals = new LinkedList();
            Face face = (Face)fIt.next();
            /*
            if (face.isBoundary()) {
                System.out.println("boundary face detected");
            }
            */
            Iterator vIt = face.vertexIterator();
            while (vIt.hasNext()) {
                VertexHandle vh = 
                    this.getVertexHandle(((Integer)vIt.next()).intValue());
                normals.add(vh.calcNormal());
                vertices.add(vh.getVertex());
            }
            pw.println("P " + vertices.size() + " 0 0");
            Iterator lIt = vertices.iterator();
            Iterator nIt = normals.iterator();

            /*
            List temp1 = new LinkedList();
            List temp2 = new LinkedList();
            */

            while (lIt.hasNext()) {
                //output the normal
                Vector3d n = (Vector3d)nIt.next();
                //temp1.add(n);
                pw.println("n " + n.x + " " +
                        n.y + " " +
                        n.z);

                //output the vertex
                Point3d v = (Point3d)lIt.next();
                //temp2.add(v);
                pw.println("v " + v.x + " " +
                        v.y + " " +
                        v.z);
            }
            pw.println("E 0 0 0");

            //****************************************
            //For DEBUG ONLY
            /*
            for (int i = 0; i < temp1.size(); i++) {
                Vector3d n = (Vector3d)temp1.get(i);
                Point3d p = (Point3d)temp2.get(i);
                pw.println("P 2 0 0");
                pw.println("v " + p.x +
                        " " + 
                        p.y + " " +
                        p.z);
                pw.println("v " + 
                        n.x + 
                        " " + 
                        n.y + " " +
                        n.z);
                pw.println("E 0 0 0");
            }
            */
            //****************************************
        }
        pw.close();
        /*
        System.out.println("There are a total of " + numFaces +
                " polygons in this mesh");
                */
    }

    public VertexHandle getVertexHandle(int vertexHashCode)
    {
        return new VertexHandle(vertexIndex, vertexToEdgeMap, 
                vertexToAttributeMap, vertexHashCode);
    }

    /**
     * Remove all vertices from Mesh that isn't used in a face
     */
    public void cleanVertices()
    {
        Set keySet = vertexIndex.keySet();
        Iterator it = keySet.iterator();

        while (it.hasNext()) {
            Object key = it.next();
            if (vertexToEdgeMap.get(key) == null) {
                vertexIndex.remove(key);
            }
        }
    }

    /**
     * Add a vertex into the mesh
     */
    /*
    public VertexHandle addVertex(Point3d v)
    {
        //return addVertex(v, new Integer(v.hashCode()));
        return addVertex(v, new String(v.toString()));
        //return addVertex(v, v);
    }
    */

    /**
     * Add a vertex into the mesh
     */
    //public VertexHandle addVertex(Point3d v, Object hashCode)
    public VertexHandle addVertex(Point3d v, int hashCode)
    {
        Integer hashCodeObj = new Integer(hashCode);
        if (vertexIndex.get(hashCodeObj) == null) {
            //System.out.println("hashCode is " + hashCodeObj.hashCode());
            vertexIndex.put(hashCodeObj, v);
        }
        else {
            //System.out.println("dup index " + v);
            //throw new RuntimeException("shit adding another vertex");
        }

        if (hashCode > largestHashCode) {
            largestHashCode = hashCode;
        }

        return new VertexHandle(vertexIndex, vertexToEdgeMap, 
                vertexToAttributeMap, hashCode);
    }

    /**
     * Add a Face into the mesh
     */
    public void addFace(VertexHandle[] v)
    {
        addFace(v, null);
    }

    /**
     * triangulates using triangle fans
     */
    private void triangularAddFace(VertexHandle[] v, Object attribute)
    {
        //System.out.println("triangular addface");
        VertexHandle[] vh = new VertexHandle[3];
        vh[0] = v[0];

        int i = 1;
        while (i < (v.length - 1)) {
            vh[1] = v[i];
            vh[2] = v[++i];
            addFace(vh, attribute);
        }
    }

    /**
     * Add a Face into the mesh
     */
    public void addFace(VertexHandle[] v, Object attribute)
    {
        class AdjacentPairs
        {
            HalfEdge e1 = null;
            HalfEdge e2 = null;
        }

        isTriangular = v.length == 3;
        if (!isTriangular && triangulate) {
            triangularAddFace(v, attribute);
            return;
        }
        /*
        System.out.println("*****************************");
        System.out.println("Adding Face ");
        for (int a = 0; a < v.length; a++) {
            System.out.println("v - " + v[a].getVertex());
        }
        System.out.println("*****************************");
        */

        HalfEdge lastHE = null;
        HalfEdge curHE = null;
        HalfEdge startHE = null;

        Face newFace = new Face(attribute);
        VertexHandle newVertex = null;
        List adjacentPairs = new LinkedList();

        for (int a = 0; a < v.length; a++) {
            int from = a; int to = (a + 1) % v.length;
            //First we create the new half edge
            curHE = new HalfEdge();
            curHE.vertexHashCode = v[to].getVertexHashCode();
            curHE.face = newFace;

            if (lastHE != null) {
                lastHE.next = curHE;
            }
            else {
                startHE = curHE;
            }

            //Now we will try to set its sym pointer by looking for an
            //adjacent edge.  An adjacent edge is defined to be an edge
            //with both of its vertices common to this edge
            //check for adjacent triangles by looking for edges going
            //from [a+1] to [a]
            Iterator oHalfEdgeIt = v[to].oHalfEdgeListIterator();
            HalfEdge adjacentEdge;
            while (oHalfEdgeIt.hasNext()) {
                adjacentEdge = (HalfEdge)oHalfEdgeIt.next();

                if (adjacentEdge.getVertexHashCode() ==
                        v[from].getVertexHashCode()) {
                    //System.out.println("FOUND ADJACENT");
                    //We've found an edge going from [a+1] to [a]
                    //set the sym pointers accordingly
                    if (adjacentEdge.sym != null) {
                        /*
                        System.out.println("from - " + 
                                v[from].getVertexHashCode() + 
                                " to - " + v[to].getVertexHashCode());
                        System.out.println("adjacentEdge is pointing from "+
                                adjacentEdge.sym.getVertexHashCode() + " to " 
                                + adjacentEdge.getVertexHashCode());
                        throw new RuntimeException("error - non " +
                                "manifold faces detected");
                        */
                        return;
                    }
                    AdjacentPairs ap = new AdjacentPairs();
                    ap.e1 = adjacentEdge;
                    ap.e2 = curHE;
                    adjacentPairs.add(ap);
                    /*
                    adjacentEdge.sym = curHE;
                    curHE.sym = adjacentEdge;
                    */
                }
            }
            lastHE = curHE;
            /*
            List edgeList = v[from].getEdgeList();
            if (edgeList == null) {
                edgeList = new LinkedList();
            }
            //Add this edge to the edge list
            edgeList.add(curHE);
            v[from].setEdgeList(edgeList);
            */
        }//for all vertices
        //complete the circular linked list
        lastHE.next = startHE;

        //We are doing these two things last because if we bail out of addFace
        //due to a offending face (non-manifold) then we don't want to leave
        //any garbage behind
        curHE = startHE;
        for (int a = 0; a < v.length; a++) {
            List edgeList = v[a].getEdgeList();
            if (edgeList == null) {
                edgeList = new LinkedList();
            }
            //Add this edge to the edge list
            edgeList.add(curHE);
            v[a].setEdgeList(edgeList);

            curHE = curHE.next;
        }
        Iterator it = adjacentPairs.iterator();
        while (it.hasNext()) {
            AdjacentPairs ap = (AdjacentPairs)it.next();
            ap.e1.sym = ap.e2;
            ap.e2.sym = ap.e1;
        }
        //Lastly set the face to point to one of the HE's
        newFace.edge = curHE;
        faces.add(newFace);
    }

    public Iterator faceIterator()
    {
        return faces.iterator();
    }

    public Iterator vertexIterator()
    {
        return new Iterator()
        {
            Iterator key = vertexIndex.keySet().iterator();

            public boolean hasNext()
            {
                return key.hasNext();
            }
            public Object next()
            {
                return key.next();
            }
            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }
}
